unit hiMST_DragRowManager;                                                                             

interface
     
uses Windows, Messages, Kol, Share, Debug;

type
  
  TIDragManager = record
    onmessage: function(Sender: PControl; var Msg: TMsg; var Rslt: Integer;
	                    fctl3d: boolean; winstyle: integer): Integer of object;
	rowdata: procedure(value: TData) of object;
	dragRow: function: boolean of object;
	getdragrowdata: function: TData of object;
	stopDrag: procedure  of object;
	enddragrow: procedure of object;
  end;
  IDragManager = ^TIDragManager;
  
  THIMST_DragRowManager = class(TDebug)
   private
	 twbm: TIDragManager;
     Bmp:PBitmap;
	 sControl,
	 rowControl: PControl;
	 bDragRow,
	 bStopDrag: boolean;
	 h,
	 step,
	 oldPos,
     cbState,
     oldIndex,
     newIndex: integer;
  	 oldCursor: HCursor;
	 dragData: TData;    

     procedure _OnPaint( Sender: PControl; DC: HDC );
     function OnMessage(Sender: PControl; var Msg: TMsg; var Rslt: Integer;
	                      fctl3d: boolean; winstyle: Integer): Integer;
     function DragRow: boolean;
     function GetDragRowData: TData;
     procedure RowData(value: TData);
     procedure StopDrag;
     procedure EndDragRow;
   public
     _prop_Name: string;
	 _prop_onEndAfterStop,
	 _prop_CursorHand: boolean;
     _data_index,
	 _event_onEscCell,
     _event_onEnterCell,
     _event_onCellSize,
     _event_onClientRect,
     _event_onVisibleCell,          
	 _event_onStart,
     _event_onDrag,
     _event_onDrop,
     _event_onEnd,
     _event_onStop: THI_Event;

	 constructor Create;
	 destructor Destroy; override;
	 function getInterfaceDragRowManager: IDragManager;
     procedure _work_doStop(var _Data: TData; Index: word);
     procedure _work_doRefresh(var _Data:TData; Index:word);     
     procedure _var_Select(var _Data:TData; Index:word);
     procedure _var_Bitmap(var _Data:TData; Index:word);
     procedure _var_Handle(var _Data:TData; Index:word);
     procedure _var_Width(var _Data:TData; Index:word);
     procedure _var_Height(var _Data:TData; Index:word);
  end;

implementation

constructor THIMST_DragRowManager.Create;
begin
  inherited;
  Bmp := NewBitmap(0,0);
  bDragRow  := false;
  bStopDrag := false;
  oldPos := 0;
  twbm.onmessage := OnMessage;
  twbm.rowdata := RowData;
  twbm.getdragrowdata := GetDragRowData;
  twbm.enddragrow := EndDragRow;
  twbm.stopdrag := StopDrag;
  twbm.dragrow := DragRow;
end;

destructor THIMST_DragRowManager.Destroy;
begin
  Bmp.Free;
  inherited;
end;

function THIMST_DragRowManager.getInterfaceDragRowManager;
begin
  Result := @twbm;
end;

function THIMST_DragRowManager.DragRow;
begin
  Result := bDragRow;
end;

procedure THIMST_DragRowManager.StopDrag;
begin
  bDragRow  := false;
  bStopDrag := true;
  newIndex  := oldIndex;
  EndDragRow;
end;

procedure THIMST_DragRowManager.RowData;
begin
  dragData := Value;
  sControl.LVDelete(oldIndex);
end;

function THIMST_DragRowManager.GetDragRowData;
var ind: integer;
    dt: TData;
begin
  ind := ReadInteger(dt,_data_Index,-1);
  if (ind >= 0) and (ind < sControl.Count) then
   newIndex := ind;
  dtNull(dt);
  dtInteger(dt, newIndex);
  dt.ldata := @dragData;
  Result := dt;
end;

procedure THIMST_DragRowManager.EndDragRow;
begin
  Bmp.Clear;
  rowControl.Visible := false;
  rowControl.Free;
  sControl.LVCurItem := newIndex;
  sControl.LVItemStateImgIdx[newIndex] := cbState;  
  sControl.LVMakeVisible(newIndex, false);
  if _prop_CursorHand then
    sControl.Cursor := oldCursor;
  dtNull(dragData);
  if bStopDrag then
   begin
    _hi_onEvent(_event_onStop,newIndex);
    if _prop_onEndAfterStop then     
      _hi_onEvent(_event_onEnd,newIndex);
   end
  else 
   _hi_onEvent(_event_onEnd,newIndex);
end;

function THIMST_DragRowManager.OnMessage;
type
  TMouseDownPt = packed record
    X: Word;
    Y: Word;
  end;

var dc:HDC;
    Pt: TMouseDownPt; 
    HTI: TLVHitTestInfo;
    b: integer;
    R: TRect;
  
  function Drag(pos: integer):integer;
  begin
   Result := -1;
   if bStopDrag then exit;
   if bDragRow then
     begin
	   rowControl.Top := pos - h;
	   _hi_onEvent(_event_onDrag);
     end
   else//noDrag
     begin
	   step := abs(oldPos - pos);
	   if step > 8 then
         begin
		   if sControl.LVSelCount > 1 then//if multiselect
		    begin
		     sControl.LVCurItem := -1;
		     sControl.LVCurItem := oldIndex;
		    end;
		   sControl.LVMakeVisible(oldIndex, false);//if scrolling fast 
		   sControl.LVItemState[oldIndex] := [lvisHighlight];
		   R := sControl.LVItemRect(oldIndex, lvipBounds);
		   h := (r.Bottom - r.Top) div 2;
		   b := 0;
		   if fctl3d then
			case winstyle of
          	  0: b := 3;
        	  1: b := 1;
        	  2: b := 2;
        	  3: b := 4;
       		  4: b := 6;
      		  end
    	    else
     		 b := 1;
		   rowControl := NewPaintbox(sControl);
 		   rowControl.Visible := False;
  		   rowControl.OnPaint := _OnPaint;
		   rowControl.Focused := true;
		   if _prop_CursorHand then
		    begin
		     oldCursor := sControl.Cursor;
		     sControl.CursorLoad(0, MakeIntResource(114));
		    end;
		   Bmp.Width := R.Right;
           Bmp.Height := R.Bottom - R.Top;
		   dc := GetWindowDC(sControl.Handle);
       	   BitBlt(Bmp.Canvas.Handle,0,0,Bmp.Width,Bmp.Height,DC,4,R.Top + 1,SRCCOPY);  
           ReleaseDC(sControl.Handle,dc);
		   SetParent(rowControl.Handle,sControl.Handle);
		   rowControl.Visible := True;
		   rowControl.Left := b;
		   rowControl.Width := min(r.Right - b,sControl.Width - GetSystemMetrics(SM_CXVSCROLL) - b * 2 - 1); 
		   rowControl.Height := R.Bottom - R.Top;
		   bDragRow := true;
		   cbState := sControl.LVItemStateImgIdx[oldIndex];
  		   oldPos := pos;
	   	   ReleaseCapture;
       	   SetCapture(sControl.Handle);       	   
		    _hi_onEvent(_event_onStart, oldIndex);
		   Result := oldIndex; 
 		 end;
     end;   
   end;

  function Drop(y:integer):integer;
  begin
    Result := -1;
	ReleaseCapture;
    if bStopDrag then
     begin
      bStopDrag := false;
      exit;
     end;
    if not bDragRow then exit;
    oldPos := y;
	if y < 0 then
	  newIndex := 0
	else
	 if y > sControl.Height then
	   newIndex := sControl.Count
     else
       begin
         newIndex := sControl.LVItemAtPos(sControl.LVColWidth[0],y);
         if (newIndex = -1) then
           newIndex := sControl.Count - 1;
		 R := sControl.LVItemRect(newIndex, lvipBounds);
      	 if (R.Top + h) < y then
           newIndex := min(sControl.Count, newIndex + 1);
       end;
    bDragRow := false;
    _hi_onEvent(_event_onDrop, newIndex);
    Result := -2;
  end;
  
  function GetPositionStructure: boolean;
   begin
    Pt:= TMouseDownPt(Msg.lParam);
    HTI.pt.x := Pt.X;
    HTI.pt.y := Pt.Y;
    sControl.Perform(LVM_SUBITEMHITTEST, 0, Integer( @HTI ));
    Result := (HTI.flags <> LVHT_ONITEMSTATEICON) and (HTI.iItem <> -1)
   end;

begin
  sControl := Sender;
  Result := -1;
  if (sControl.LVStyle = lvsDetail) or (sControl.LVStyle = lvsDetailNoHeader) then
   case Msg.message of
    WM_LBUTTONDOWN:
     begin
      if GetPositionStructure then
       begin 
        oldIndex := HTI.iItem;
        step := 0;
        oldPos := HTI.pt.y;
       end;
     end;
    WM_RBUTTONDOWN:
	 if bDragRow then
	   Result := Drop(SmallInt(hiWord(Msg.lParam)));
    WM_LBUTTONUP:
	  Result := Drop(SmallInt(hiWord(Msg.lParam)));
	WM_MOUSEMOVE:
     begin
   	  if not bDragRow and (sControl.Count < 2) then exit;
       case (Msg.wParam and MK_LBUTTON) of
        0: oldPos := SmallInt(hiWord(Msg.lParam));
	    1: Result := Drag(SmallInt(hiWord(Msg.lParam)));
	   end;
	 end; 
   end;
end;

procedure THIMST_DragRowManager._OnPaint;
begin
 if not Bmp.Empty then
  Bmp.Draw(DC,0,0);
end;

procedure THIMST_DragRowManager._work_doStop;
begin
  if bDragRow then
   begin
    bDragRow  := false;
    bStopDrag := true;
    newIndex  := oldIndex;
    PostMessage(sControl.Handle, US_STOPDRAG, 0, 0);
   end;
end;

procedure THIMST_DragRowManager._work_doRefresh;
begin
  rowControl.Invalidate;
end;

procedure THIMST_DragRowManager._var_Select;
begin
  if bDragRow then
   _Data := dragData
  else
   dtNull(_Data);   
end;

procedure THIMST_DragRowManager._var_Bitmap;
begin
   dtBitmap(_data,Bmp);
end;

procedure THIMST_DragRowManager._var_Handle;
begin
 if bDragRow then
  dtInteger(_data, rowControl.GetWindowHandle)
 else
  dtInteger(_data, 0);
end;

procedure THIMST_DragRowManager._var_Width; 
begin 
 if bDragRow then
  dtInteger(_data, rowControl.Width)
 else
  dtInteger(_data, 0);
end;

procedure THIMST_DragRowManager._var_Height; 
begin 
 if bDragRow then
  dtInteger(_data, rowControl.Height)
 else
  dtInteger(_data, 0);
end; 

end.